/* Forth VM in java
 * started 31.12.2006
 */

import java.lang.*;
import java.util.*;
import ral.*;

public class fvm {
    /* --- global data areas --- */
    private Real[] ds = new Real[50]; /* Data stack */
    private int[] rs = new int[100]; /* Return stack */
    private loop[] ls = new loop[20]; /* Loop stack */
    private Vector code = new Vector(); /* code & data */
    private Real[] vars = new Real[1000];
    private Hashtable voc = new Hashtable(); /* vocabulary */
    private Hashtable vocMacro = new Hashtable(); /* immediate words */
    public String str; /* input string */

    /* --- global variables --- */
    private int dp=0;
    private int state=1; /* interpret by default */
    private int ip=1; 		/* ip of code */
    public int dsp=-1;		/* data stack pointer */
    public int rsp=-1;		/* return stack pointer */
    public int lsp=-1;		/* loop stack pointer */
    int sp=0; // string pointer
    Real.NumberFormat format=new Real.NumberFormat();
    String last_word;	/* last defined word */
    int last_word_adr;
    String tmp_wrd;
    Integer tmp_int;
    boolean bad_word=false;
    public boolean quit=false;
    
    /* --- stuff for plotting graphs --- */
    private Real pltA=new Real(0);
    private Real pltB=new Real(0);
    public int ySize=200;
    public int xSize=100;
    public Real[][] graphs; /* max 10 plots xSize bins each */
    public int[][] gr;
    public int numberOfGraphs=0;
    private Real tmpvar = new Real(0);
    private Real tmpvar2 = new Real(0);
    private Real gmax=new Real(0);
    private Real gmin=new Real(0);
    public boolean in_graph=false;
    
    public fvm() {
	
	initvoc();
    }	
    
    
    /* --- The execution core --- */
    
    public void exec(int val) throws RuntimeException {  /* execute one instruction */
	if (val>=0)
	{
	    rsp++;
	    rs[rsp]=ip;
	    ip=val;
	}
	else
	{
	    if(val==-1) {   /* return */
		ip=rs[rsp];
		rsp--;
	}
	    switch(val)
		{
		    case -2: /* quit */
		    {
		    rsp++;
		    rs[rsp]=ip;
		    quit=true;
		    throw new RuntimeException();
		    } 
		    case -3: /* dup */
		    {
		    ds[dsp+1].assign(ds[dsp]);
		    dsp++;
		    } break;
		    case -4: /* drop *****/
		    {
		    dsp--;
		    } break;
		    case -5: /* lit *****/
		    {
		    dsp++;
		    ds[dsp].assign((Real)code.elementAt(ip));
		    ip++;
		    } break;
		    case -6: /* . dot print (debug) *****/
		    {
		    output(((Real)ds[dsp]).toString(format)+" ");
//		    output(" ");
		    dsp--;
		    } break;
		    case -7: /* swap *****/
		    {
		    ds[dsp-1].swap(ds[dsp]);
		    } break;
		    case -8: /* rot  a b c -- c a b *****/
		    {
		    tmpvar.assign(ds[dsp-2]);
		    ds[dsp-2].assign(ds[dsp]);
		    tmpvar.swap(ds[dsp-1]);
		    ds[dsp].assign(tmpvar);
		    } break;
		    case -9: /* + */
		    {
		    ds[dsp-1].add(ds[dsp]);
		    dsp--;
		    } break;
		    case -10: /* - */
		    {
		    ds[dsp-1].sub(ds[dsp]);
		    dsp--;
		    } break;
		    case -11: /* * */
		    {
		    ds[dsp-1].mul(ds[dsp]);
		    dsp--;
		    } break;
		    case -12: /* / */
		    {
		    ds[dsp-1].div(ds[dsp]);
		    dsp--;
		    } break;
		    case -13: /* abs */
		    {
		    ds[dsp].abs();
		    } break;
		    case -14: /* exp */
		    {
		    ds[dsp].exp();
		    } break;
		    case -15: /* sqrt */
		    {
		    ds[dsp].sqrt();
		    } break;
		    case -16: /* sin */
		    {
		    ds[dsp].sin();
		    } break;
		    case -17: /* asin */
		    {
		    ds[dsp].asin();
		    } break;
		    case -18: /* cos */
		    {
		    ds[dsp].cos();
		    } break;
		    case -19: /* acos */
		    {
		    ds[dsp].acos();
		    } break;
		    case -20: /* tan */
		    {
		    ds[dsp].tan();
		    } break;
		    case -21: /* atan */
		    {
		    ds[dsp].atan();
		    } break;
		    case -22: /* pi */
		    {
		    dsp++;
		    ds[dsp].assign(Real.PI);
		    } break;
		    case -23: /* @ */
		    {
		    ds[dsp].assign((Real)code.elementAt(ds[dsp].toInteger()));
		    } break;
		    case -24: /* ! */
		    {
		    ((Real)code.elementAt(ds[dsp].toInteger())).assign(ds[dsp-1]);
		    dsp-=2;
		    } break;
		    case -25: /* {   bound start-count */
		    {
		    lsp++;
		    ls[lsp].count.assign(ds[dsp]);
		    ls[lsp].bound.assign(ds[dsp-1]);
		    ls[lsp].addr=ip;
		    ls[lsp].delta.assign(Real.ONE);
		    ls[lsp].adj();
		    dsp-=2;
		    } break;
		    case -26: /* } */
		    {
		    ls[lsp].incr();
		    ip=ls[lsp].addr;
		    } break;
		    case -27: { /* i */
			dsp++;
			ds[dsp].assign(ls[lsp].count);
		    } break;
		    case -28: { /* j */
			dsp++;
			ds[dsp].assign(ls[lsp-1].count);
		    } break;
		    case -29: { /* k */
			dsp++;
			ds[dsp].assign(ls[lsp-2].count);
		    } break;
		    case -30: /* 0 */
		    {
		    dsp++;
		    ds[dsp].assign(Real.ZERO);
		    } break;
		    case -31: /* space */
		    {
		    output(" ");
		    } break;
		    case -32: /* cr */
		    {
		    output("\n");
		    } break;
		    case -33: /* +{ */
		    {
		    lsp++;
		    ls[lsp].count.assign(ds[dsp-1]);
		    ls[lsp].bound.assign(ds[dsp-2]);
		    ls[lsp].addr=ip;
		    ls[lsp].delta.assign(ds[dsp]);
		    ls[lsp].adj();
		    dsp-=3;
		    } break;
//		    case -33: /* +} */
//		    {
//		    ls[lsp].incr(ds[dsp]);
//		    dsp--;
//		    ip=ls[lsp].addr;
//		    } break;
		    case -34: /* ln */
		    {
		    ds[dsp].ln();
		    } break;
		    case -35: /* lg */
		    {
		    ds[dsp].log10();
		    } break;
		    case -36: /* ^ */
		    {
		    ds[dsp-1].pow(ds[dsp]);
		    dsp--;
		    } break;
		    case -37: /* max */
		    {
		    if(ds[dsp].greaterThan(ds[dsp-1])) ds[dsp-1].assign(ds[dsp]);
		    dsp--;
		    } break;
		    case -38: /* min */
		    {
		    if(ds[dsp-1].greaterThan(ds[dsp])) ds[dsp-1].assign(ds[dsp]);
		    dsp--;
		    } break;
		    case -39: /* rnd */
		    {
		    dsp++;
		    ds[dsp].random();
		    } break;
		    case -40: /* erfc */
		    {
		    ds[dsp].erfc();
		    } break;
		    case -42: /* = */
		    {
		    dsp--;
		    if(ds[dsp].equalTo(ds[dsp+1])) ds[dsp].assign(Real.ONE);
		    else ds[dsp].assign(Real.ZERO);
		    } break;
		    case -43: /* < */
		    {
		    dsp--;
		    if(ds[dsp].lessThan(ds[dsp+1])) ds[dsp].assign(Real.ONE);
		    else ds[dsp].assign(Real.ZERO);
		    } break;
		    case -44: /* > */
		    {
		    dsp--;
		    if(ds[dsp].greaterThan(ds[dsp+1])) ds[dsp].assign(Real.ONE);
		    else ds[dsp].assign(Real.ZERO);
		    } break;
		    case -45: /* <= */
		    {
		    dsp--;
		    if(ds[dsp].lessEqual(ds[dsp+1])) ds[dsp].assign(Real.ONE);
		    else ds[dsp].assign(Real.ZERO);
		    } break;
		    case -46: /* >= */
		    {
		    dsp--;
		    if(ds[dsp].greaterEqual(ds[dsp+1])) ds[dsp].assign(Real.ONE);
		    else ds[dsp].assign(Real.ZERO);
		    } break;
		    case -47: /* sinh */
		    {
		    ds[dsp].sinh();
		    } break;
		    case -48: /* cosh */
		    {
		    ds[dsp].cosh();
		    } break;
		    case -49: /* tanh */
		    {
		    ds[dsp].tanh();
		    } break;
		    case -50: /* asinh */
		    {
		    ds[dsp].asinh();
		    } break;
		    case -51: /* acosh */
		    {
		    ds[dsp].acosh();
		    } break;
		    case -52: /* atanh */
		    {
		    ds[dsp].atanh();
		    } break;
		    case -53: /* if, [ */
		    {
		    if(ds[dsp].isZero()) {	/* false branch */
			ip=((Integer)code.elementAt(ip)).intValue();
			} 
		    else ip++; 			/* true branch */
		    dsp--;
		    } break;
		    case -54: /* else. This is just GOTO. */
		    {
		    ip=((Integer)code.elementAt(ip)).intValue();
		    } break;
		    case -55: /* { do-loop: body */
		    {
		    if(ls[lsp].check()) {
			ip=((Integer)code.elementAt(ip)).intValue();
			lsp--;
			}
		    else {
			ip++; 
			// dsp++;
			// ds[dsp].assign(ls[lsp].count);
			}
		    } break;
		    case -66: /* over */
		    {
		    dsp++;
		    ds[dsp].assign(ds[dsp-2]);
		    } break;
		    
		    case -67: { /* graph */
			pltB.assign(ds[dsp-1]);
			pltA.assign(ds[dsp]);
			pltB.sub(pltA);
			pltB.div(xSize);
			numberOfGraphs=0;
			dsp-=2;
			in_graph=true;
			preplot();
		    } break;
		    case -68: { /* plot */
			int i;
			int func=((Integer)code.elementAt(ip)).intValue();
			Real X=new Real();
			for(i=0; i<xSize; i++) {
			    X.assign(i);
			    X.mul(pltB);
			    X.add(pltA);
			    dsp++;
			    ds[dsp].assign(X);
			    fexec(func);
			    graphs[numberOfGraphs][i].assign(ds[dsp]);
			    dsp--;
			}
			ip++;
			numberOfGraphs++;
			in_graph=true;
		    } break;
		    case -69: { /* show */
			int i; int j; 
			gmax.assign(graphs[0][1]);
			gmin.assign(graphs[0][1]);
			for(i=0;i<xSize;i++) for(j=0;j<numberOfGraphs;j++) {
			    if(gmax.lessThan(graphs[j][i])) 
				    gmax.assign(graphs[j][i]);
			    if(gmin.greaterThan(graphs[j][i])) 
				    gmin.assign(graphs[j][i]);
			}
			gmax.sub(gmin);
			for(i=0;i<xSize;i++) for(j=0;j<numberOfGraphs;j++) {
			    graphs[j][i].sub(gmin);
			    graphs[j][i].div(gmax);
			    graphs[j][i].mul(ySize);
			    gr[j][i]=graphs[j][i].toInteger();
			}
			plot();
			in_graph=false;
			numberOfGraphs=0;
		    } break;
		    case -70: { /* yshow */
			int i; int j; 
			gmax.assign(ds[dsp]);
			gmin.assign(ds[dsp-1]);
			dsp-=2;
			gmax.sub(gmin);
			for(i=0;i<xSize;i++) for(j=0;j<numberOfGraphs;j++) {
			    graphs[j][i].sub(gmin);
			    graphs[j][i].div(gmax);
			    graphs[j][i].mul(ySize);
			    gr[j][i]=graphs[j][i].toInteger();
			}
			plot();
			in_graph=false;
			numberOfGraphs=0;
		    } break;
		    case -71: { /* sleep */
			sleeep(ds[dsp].toLong());
			dsp--;
		    } break;
		    case -72: { /* int */
			ds[dsp].floor();
		    } break;
		    
		    case -73: { /* (to) */
			((Real)code.elementAt(((Integer)code.elementAt(ip)).intValue())).assign(ds[dsp]);
			dsp--;
    			ip++;
		    } break;
		    
		    case -74: { /* (from) */
			dsp++;
			ds[dsp].assign((Real)code.elementAt(ip));
			ip=rs[rsp];
			rsp--;
		    } break;
		    
		    case -76: { /* var */
			aw(word(32).toLowerCase(),here());
			compile_int(-74);
			compile(new Real(Real.ZERO));
		    } break;
		    
		    case -79: { /* ['] */
			dsp++;
			ds[dsp].assign( ((Integer)code.elementAt(ip)).intValue() );
			ip++;
		    } break;
		    case -80: { /* exec */
			dsp--;
			fexec( (ds[dsp+1]).toInteger() );
		    } break;
		    case -81: {  /* self */
			dsp++;
			ds[dsp].assign(ip);
			ip=rs[rsp];
			rsp--;
		    } break; 
		    case -82: { /* self-in-does */
			dsp++;
			ds[dsp].assign(rs[rsp]);
			rsp--;
		    } break;

		    case -83: { /* r> */
			dsp++;
			ds[dsp].assign(rs[rsp]);
			rsp--;
		    } break;
		    
		    case -84: { /* >r */
			rsp++;
			ds[dsp].round();
			rs[rsp]=ds[dsp].toInteger();
			dsp--;
		    } break;
		    
		    case -85: { /* not */
			if(ds[dsp].equalTo(Real.ZERO)) ds[dsp].assign(Real.ONE);
			else ds[dsp].assign(Real.ZERO); 
		    } break;
		    
		    case -86: { /* or */
			dsp--;
			if(ds[dsp].equalTo(Real.ZERO)&&ds[dsp+1].equalTo(Real.ZERO))
			    ds[dsp].assign(Real.ZERO);
			else ds[dsp].assign(Real.ONE);
		    } break;
		    
		    case -87: { /* and */
			dsp--;
			if(ds[dsp].equalTo(Real.ZERO)||ds[dsp+1].equalTo(Real.ZERO))
			    ds[dsp].assign(Real.ZERO);
			else ds[dsp].assign(Real.ONE);
		    } break;
		    
		    case -88: { /* 1+ */
			ds[dsp].add(Real.ONE);
		    } break;
		    
		    case -89: { /* 1- */
			ds[dsp].sub(Real.ONE);
		    } break;
		    
		    case -90: { /* 2* */
			ds[dsp].scalbn(1);
		    } break;
		    
		    case -91: { /* 2/ */
			ds[dsp].scalbn(-1);
		    } break;
		    
		    case -92: { /* % */
			ds[dsp].recip();
		    } break;
		
		    case -93: { /* ." */
			output((String)code.elementAt(ip));
			ip++;
		    } break;
		    
		    case -94: { /* r@ */
			dsp++;
			ds[dsp].assign(rs[rsp]);
		    } break;
		    
		    case -95: { /* step */
			dsp++;
			ds[dsp].assign(ls[lsp].delta);
		    } break;
		    
		    case -96: { /* quit0 */ // like quit, but rs points to itself
			rsp++;              // needed for empty input strings
			rs[rsp]=0;
			ip=code.size();
		    } break;
		    case -97: { /* bye */
			bye();
		    } break;
		    
		    case -98: { /* format */
			format.maxwidth=ds[dsp].toInteger();
			dsp--;
		    } break;
		    
		    case -99: { /* depth */
			dsp++;
			ds[dsp].assign(dsp);
		    } break;
		    
		    case -100: { /* cli */ // clear input
			cli();
		    } break;
		    case -101: { /* clo */ // clear input
			clo();
		    } break;
		    
		    case -102: { /* milliseconds */ //show seconds since 1970
			dsp++;
			ds[dsp].assign((new Date()).getTime());
		    } break;
		    case -103: { /* randomize */
			Real.accumulateRandomness(ds[dsp].toLong());
			dsp--;
		    } break;
		    case -104: { /* today */
			dsp+=3;
			ds[dsp-2].assign(Calendar.getInstance().get(Calendar.YEAR));
			ds[dsp-1].assign(Calendar.getInstance().get(Calendar.MONTH)+1);
			ds[dsp].assign(Calendar.getInstance().get(Calendar.DATE));
		    } break;
		    case -105: { /* now */
			dsp+=3;
			ds[dsp-2].assign(Calendar.getInstance().get(Calendar.HOUR_OF_DAY));
			ds[dsp-1].assign(Calendar.getInstance().get(Calendar.MINUTE));
			ds[dsp].assign(Calendar.getInstance().get(Calendar.SECOND));
		    } break;
		    
		    case -106: { /* norm */
			format.fse=format.FSE_NONE;
		    } break;
		    case -107: { /* sci */
			format.fse=format.FSE_SCI;
		    } break;
		    case -108: { /* eng */
			format.fse=format.FSE_ENG;
		    } break;
		    
		    case -109: { /* here */
			dsp++;
			ds[dsp].assign(code.size());
		    } break;
		    case -110: { /* allot */
			for(int i=0;i<ds[dsp].toInteger();i++)
			    code.addElement(new Real(Real.ZERO));
			dsp--;
		    } break;
		    
		    case -111: { /* create */
			last_word=word(32).toLowerCase();
			last_word_adr=here();
			aw(last_word,here());
			compile_int(-81); // return address
		    } break;
		    case -112: { /* (does>) */
			code.setElementAt(new Integer(ip),last_word_adr);
			ip=rs[rsp];
			rsp--;
		    } break;
		    case -113: { /* , */
			compile(new Real(ds[dsp]));
			dsp--;
		    } break;
		    case -114: { /* ' */
			dsp++;
			ds[dsp].assign(get_word_adr());
		    } break;
		    case -115: { /* immediate */
			awim(last_word,last_word_adr);
		    } break;
		    case -116: { /* last_word */
			dsp++;
			ds[dsp].assign(last_word_adr);
		    } break;
		    case -117: { /* i@ */
			ds[dsp].assign(((Integer)code.elementAt(ds[dsp].toInteger())).intValue());
		    } break;
		    case -118: { /* i! */
			code.setElementAt(new Integer(ds[dsp-1].toInteger()),ds[dsp].toInteger());
			dsp-=2;
		    } break;
		    case -119: { /* c! */
			code.setElementAt(new Character((char)(ds[dsp-1].toInteger())),ds[dsp].toInteger());
			dsp-=2;
		    } break;
		    case -120: { /* c@ */
			ds[dsp].assign( (int)( ((Character)code.elementAt(ds[dsp].toInteger())).charValue() ) );
		    } break;
		    case -121: { /* emit */
			output( (new Character((char)(ds[dsp].toInteger()))).toString() );
			dsp--;
		    } break;
		    
		    case -129: { /* line */
			line(ds[dsp-1].toInteger(),ds[dsp].toInteger(),1);
			dsp-=4;
		    } break;
		    case -130: { /* color */
			setColor(ds[dsp].toInteger());
			dsp--;
		    } break;
		    case -131: { /* box */
			line(ds[dsp-1].toInteger(),ds[dsp].toInteger(),2);
			dsp-=2;
		    } break; 
		    case -132: { /* pixel */
			pixel(ds[dsp-1].toInteger(),ds[dsp].toInteger());
			dsp-=2;
		    } break;
		    case -133: { /* ?key */
			dsp++;
			ds[dsp].assign(getKey());
		    } break;
		    case -134: { /* xsize */
			dsp++;
			ds[dsp].assign(xSize);
		    } break;
		    case -135: { /* ysize */
			dsp++;
			ds[dsp].assign(ySize);
		    } break;
		    
		    
		    case -500: /* dump */
		    {
		    output(code.toString());
		    } break;
		    
		    // Immediate words
		    
		    case -501: { /* : */
			state=0;
			last_word=word(32).toLowerCase();
			last_word_adr=here();
			aw(last_word,here());
		    } break;		    
		    case -502: { /* ; */
			state=1;
			compile_int(-1);
		    } break;

		    case -503: { /* ( */
			word(')');
		    } break;
		    case -504: { /* ." */
			if(state==1) output(word('"'));
			else {
			    compile_int(-93);
			    compile_String(word('"'));
			}
		    } break;
		    
		    case -505: { /* if */
			startControl(-53);
		    } break;
		    case -506: { /* then */
			endControl(0);
		    } break;
		    case -507: { /* else */
			endControl(2);
			startControl(-54);
		    } break;
		    
		    case -508: { /* [ */
			state=1;
		    } break;
		    case -509: { /* ] */
			state=0;
		    } break;
		    
		    case -510: { /* [COMPILE] */
			compile_int(get_word_adr());
		    } break;
		    
		    case -511: { /* begin */
			rsp++;
			rs[rsp]=here();
		    } break;
		    case -512: { /* until */
			compile_int(-53);
			compile_int(rs[rsp]);
			rsp--;
		    } break;
		    case -513: { /* do */
			compile_int(-25);
			startControl(-55);
		    } break;
		    case -514: { /* loop */
			compile_int(-26);
			endControl(0);
		    } break;
		    
		    case -515: { /* to */
			//int adr=((Integer)voc.get(word(32).toLowerCase())).intValue()+1;
			if(state==1) {
			    ((Real)code.elementAt(get_word_adr()+1)).assign(ds[dsp]);
			    dsp--;
			} else {
			    compile_int(-73); // (to)
			    compile_int(get_word_adr()+1);
			}
		    } break;
		    
		    case -516: { /* does */
			compile_int(-112); // (does>)
			compile_int(-82); // self-in-does
		    } break;
		    
		    case -517: { /* +do */
			compile_int(-33);
			startControl(-55);
		    } break;
		    case -518: { /* repeat */
			endControl(2);
			compile_int(-54);
			compile_int(rs[rsp]);
			rsp--;
		    } break;
		    
		    case -519: { /* S" */
			String tmpstr=word('"');
			if(state==1) {
			    dsp++;
			    ds[dsp].assign(here());
			    for(int i=0;i<tmpstr.length();i++) {
				code.addElement(new Character(tmpstr.charAt(i)));
			    }
			    dsp++;
			    ds[dsp].assign(tmpstr.length());
			} else {
			    compile_int(-5);
			    compile(new Real(here()+3));
			    startControl(-54);
			    for(int i=0;i<tmpstr.length();i++) {
				code.addElement(new Character(tmpstr.charAt(i)));
			    }
			    endControl(0);
			    compile_int(-5);
			    compile(new Real(tmpstr.length()));
			}
		    } break;
		    
		    
		} /* end switch */
	    } /* end if */
        }

    /* --- The vocabulary --- */

    void aw(String wrd, int val) { /* add word */
	voc.put(wrd,new Integer(val));
    }
    void awim(String wrd, int val) { /* add immediate word */
	vocMacro.put(wrd,new Integer(val));
    }
    
    public void initvoc() {
	graphs = new Real[10][xSize];
	gr = new int[10][xSize];
	format.maxwidth=8;
	state=1;
//	Real.magicRounding=false;
//	code.addElement(new Integer(-96)); // quit0 at the beginning of code!
//	rsp++; rs[rsp]=0;
	for(int i=0;i<1000;i++) vars[i]=new Real(0);
	for(int i=0;i<50;i++) ds[i]=new Real(0);
	for(int i=0;i<100;i++) rs[i]=0;
	for(int i=0;i<20;i++) ls[i]=new loop();
	for(int i=0;i<10;i++) for(int j=0;j<xSize;j++) graphs[i][j]=new Real();
	aw("ret",-1);
	aw("quit",-2);
	aw("dup",-3);
	aw("drop",-4);
	aw(".",-6);
	aw("swap",-7);
	aw("rot",-8);
	aw("+",-9);
	aw("-",-10);
	aw("*",-11);
	aw("/",-12);
	aw("abs",-13);
	aw("exp",-14);
	aw("sqrt",-15);
	aw("sin",-16);
	aw("asin",-17);
	aw("cos",-18);
	aw("acos",-19);
	aw("tan",-20);
	aw("atan",-21);
	aw("pi",-22);
	aw("@",-23);
	aw("!",-24);
	aw("i",-27);
	aw("j",-28);
	aw("k",-29);


	aw("0",-30);
	aw("space",-31);
	aw("cr",-32);

	aw("ln",-34);
	aw("lg",-35);
	aw("^",-36);
	aw("max",-37);
	aw("min",-38);
	aw("rnd",-39);
	aw("erfc",-40);
	aw("=",-42);
	aw("<",-43);
	aw(">",-44);
	aw("<=",-45);
	aw(">=",-46);
	aw("sinh",-47);
	aw("cosh",-48);
	aw("tanh",-49);
	aw("asinh",-50);
	aw("acosh",-51);
	aw("atanh",-52);
	
	aw("over",-66);
	
	aw("graph",-67);
	aw("plot",-68);
	aw("show",-69);
	aw("yshow",-70);
	aw("sleep",-71);
	aw("int",-72);
	
	aw("val",-76);
	aw("value",-76);
	
	aw("[']",-79);
	aw("exec",-80);
	
	aw("r>",-83);
	aw(">r",-84);
	aw("not",-85);
	aw("or",-86);
	aw("and",-87);
	aw("1+",-88);
	aw("1-",-89);
	aw("2*",-90);
	aw("2/",-91);
	aw("%",-92);
	aw("r@",-94);
	aw("step",-95);
	aw("bye",-97);
	aw("format",-98);
	aw("depth",-99);
	aw("cli",-100);
	aw("clo",-101);
	aw("milliseconds",-102);
	aw("randomize",-103);
	aw("today",-104);
	aw("now",-105);
	aw("norm",-106);
	aw("sci",-107);
	aw("eng",-108);
	aw("here",-109);
	aw("allot",-110);
	aw("create",-111);
	aw(",",-113);
	aw("\'",-114);
	aw("immediate",-115);
	aw("last_word",-116);
	aw("i@",-117);
	aw("i!",-118);
	aw("c!",-119);
	aw("c@",-120);
	aw("emit",-121);
	aw("line",-129);
	aw("color",-130);
	aw("box",-131);
	aw("pixel",-132);
	aw("?key",-133);
	aw("xsize",-134);
	aw("ysize",-135);
	
	aw("dump",-500);
	
	// immediate words
	
	awim(":",-501);
	awim(";",-502);
	awim("(",-503);
	awim(".\"",-504);
	awim("if",-505);
	awim("then",-506);
	awim("else",-507);
	awim("[",-508);
	awim("]",-509);
	awim("[compile]",-510);
	awim("begin",-511);
	awim("until",-512);
	awim("do",-513);
	awim("loop",-514);
	awim("to",-515);
	awim("does>",-516);
	awim("+do",-517);
	awim("repeat",-518);
	awim("while",-505);
	awim("s\"",-519);
	
    }
    
    int get_word_adr() throws RuntimeException {
	String wrd=word(32).toLowerCase();	
	Integer tmp=(Integer)vocMacro.get(wrd);
	if(tmp==null) tmp=(Integer)voc.get(wrd);
	if(tmp==null) { 
	    bad_word=true;
	    last_word=wrd;
	    throw new RuntimeException();
	}
	return(tmp.intValue());
    }
    
    
    /* --- The interpreter (&compiller) --- */
    
    void fexec(int vv) {  // execution loop
	exec(vv);
	if(vv<0) return;
	
	int val;
	int old_rsp=rsp;
	while(rsp>=old_rsp){

	    val = ((Integer)code.elementAt(ip)).intValue();
	    ip++;   /* ip points the next element */
	    exec(val);
	} /* end while */
    } /* end fexec */
    
    private void startControl(int val) {  // compile word val, reserve space for address 
	code.addElement(new Integer(val));
	rsp++;
	rs[rsp]=code.size();
	code.addElement(new Integer(0));
    }
    
    private void endControl(int addrInc) { // fill reserved space
	/* addrInc is the number of additional cells to skip */
	code.setElementAt(new Integer(code.size()+addrInc),rs[rsp]);
	rsp--;
    }

    private String word(int ch) { // find and return end of next word
	int spn=sp+1;
	int spnn=sp;
	do {
	    if(spn<=sp) sp++;
	    spn=str.indexOf(ch,sp);
	    spnn=str.indexOf("\n",sp);
	    if(spn<0) spn=spnn;
	    if((spnn>0) && (spnn<spn)) spn=spnn;
	    if (spn<0) spn=str.length();
	    spnn=sp;
	}while(spn<=sp);
	sp=spn+1;
	return(str.substring(spnn,spn));
    }

    void compile_int(int val) { code.addElement(new Integer(val)); }
    void compile(Object o) { code.addElement(o); }
    void compile_String(String s) { code.addElement(s); }
    int here() {return(code.size());}

    void interpret() 
    throws RuntimeException {
	int last=code.size();
	quit=false;
	sp=0;
	Integer tmp=new Integer(0);
	Integer tmpim=new Integer(0);;
	Real rtmp=new Real();
	str=str.trim();
	String wrd=new String();
	String wrd1=new String();
    
	while(sp<str.length()) {
	    wrd=word(32);
	    wrd1=wrd.toLowerCase();
	    tmp=(Integer)voc.get(wrd1);
	    tmpim=(Integer)vocMacro.get(wrd1);
	    if(tmpim!=null) { /* an immediate word */
		fexec(tmpim.intValue());
		continue;
	    }
	    if(state==1) /*interpret */ {
		if(tmp!=null)    // if it is a word, execute it
		    fexec(tmp.intValue());
		else { // a number
		    rtmp=new Real(wrd); 
		    if(rtmp.equalTo(Real.ZERO)) {
			bad_word=true;
			last_word=wrd;
			throw new RuntimeException();
		    } else {
			dsp++;
			ds[dsp].assign(rtmp);
		    }
		}
	    } else { /* compile */
		if(tmp!=null)    // if it is a word, compile it
		    compile_int(tmp.intValue()); 
		else {
		    rtmp=new Real(wrd); 
		    if(rtmp.equalTo(0)) {
			bad_word=true;
			last_word=wrd;
			throw new RuntimeException();
		    }
		compile_int(-5); /* lit */ 
		compile(rtmp);
		}
	    }
	}
	int val;
	if(rsp>-1) { ip=rs[rsp]; rsp--; }
	while(rsp>-1) {      /* this is need to continue the execution after quit */
	    val = ((Integer)code.elementAt(ip)).intValue();
	    ip++;   /* ip points the next element */
	    exec(val);
	}
    }
    
    /* --- System actions --- */
    
    public void clear() {
	code.removeAllElements();
	ip=0;
	dsp=-1;
	rsp=-1;
	lsp=-1;
	state=1;
	in_graph=false;
//	code.addElement(new Integer(-96)); // quit0 at the beginning of code!
//	rsp++; rs[rsp]=0;
    }

    public void softreset() {
	dsp=-1;
	rsp=-1;
	lsp=-1;
	ip=code.size();
    }
    
    public String diagnose() {
	String s="Error. ";
	if(dsp<-1) s=s+"Stack underflow. ";
	if(dsp>=50) s=s+"Stack overflow. ";
	if(rsp<-1) s=s+"Rstack underflow. ";
	if(rsp>=100) s=s+"Rstack overflow. ";
	if(bad_word) {
	    s=s+last_word+" not found.";
	    bad_word=false;
	}
//	System.err.println(code.toString());
	return(s);
    }

    /* --- methods to be replaced during inheritance --- */

    public void output(String s) {
	System.out.println(s);
    }
    
    public void abort(String s) {
	System.err.println(s);
    }
    
    
    public String input() {
	return(new String(""));
    }
    
    public void plot() { System.out.println("PLOT"); }
    public void preplot() {return;}
    public void save(int scr) { System.out.println("SAVE"); }
    public void load(int scr) { System.out.println("LOAD"); }

    public void sleeep(long t) { return; }
    public void cli() { return; }
    public void clo() { return; }
    
    public void bye() {return;}
    
    public void line(int x,int y,int type) {return;} // type 1=line, 2=box
    public void pixel(int x,int y) {return;}
    public void setColor(int col) {return;}
    
    public int getKey() {return(0);}
    

} /* end fvm */
